# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines ModelTask for backcompat by reimporting from new location."""

from raiutils.models import ModelTask

__all__ = ["ModelTask"]
